--[[
 Copyright (c) 2012 Robin Schoonover

 Permission is hereby granted, free of charge, to any person obtaining a copy
 of this software and associated documentation files (the "Software"), to
 deal in the Software without restriction, including without limitation the
 rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
 sell copies of the Software, and to permit persons to whom the Software is
 furnished to do so, subject to the following conditions:

 The above copyright notice and this permission notice shall be included in
 all copies or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
 IN THE SOFTWARE.
]]

local MAJOR, MINOR = "LibPetJournal-1.0", 4
local lib, oldminor = LibStub:NewLibrary(MAJOR, MINOR)

if not lib then return end

--
-- GLOBALS: PetJournal
--

local _G = _G
local hooksecurefunc, tinsert, pairs, wipe = _G.hooksecurefunc, _G.table.insert, _G.pairs, _G.wipe
local C_PetJournal, InCombatLockdown = _G.C_PetJournal, _G.InCombatLockdown

local start_background

--
--
--

lib.callbacks = lib.callbacks or LibStub("CallbackHandler-1.0"):New(lib)
lib.Pets = lib.Pets or {}
lib.UnknownPets = lib.UnknownPets or {}
lib.PetsById = lib.PetsById or {}
lib.NumPets = lib.NumPets or 0
lib.NumUnknownPets = lib.NumUnknownPets or 0

if not C_PetJournal then return end

--
-- filter anti-response
--

local filter_changed = false
for i,name in ipairs{"SetSearchFilter", "ClearSearchFilter", "SetFlagFilter",
        "SetPetSourceFilter", "SetPetTypeFilter", "AddAllPetSourcesFilter",
        "AddAllPetTypesFilter", "ClearAllPetSourcesFilter", "ClearAllPetTypesFilter"} do
    hooksecurefunc(C_PetJournal, name, function(string)
        filter_changed = true
    end)
end

--
-- filter handling
--

local filter_clear_and_save, filter_load_and_restore
do
    local PJ_FLAG_FILTERS = {
        [LE_PET_JOURNAL_FLAG_COLLECTED] = true,
        [LE_PET_JOURNAL_FLAG_NOT_COLLECTED] = true,
        [LE_PET_JOURNAL_FLAG_FAVORITES] = false
    }
    
    local function pet_journal_freeze()
        if PetJournal then
            PetJournal:UnregisterEvent("PET_JOURNAL_LIST_UPDATE")
        end
    end
    
    local function pet_journal_thaw()
        if PetJournal then
            PetJournal:RegisterEvent("PET_JOURNAL_LIST_UPDATE")
        end
    end
    
    local last_search_filter
    hooksecurefunc(C_PetJournal, "SetSearchFilter", function(string)
        last_search_filter = string
    end)
    
    hooksecurefunc(C_PetJournal, "ClearSearchFilter", function()
        last_search_filter = nil
    end)
    
    local s_search_filter
    local flag_filters = {}
    local type_filters = {}
    local source_filters = {}
    function filter_clear_and_save(search_filter)
        pet_journal_freeze()
        
        s_search_filter = last_search_filter
        C_PetJournal.ClearSearchFilter()
        
        for flag, value in pairs(PJ_FLAG_FILTERS) do
            flag_filters[flag] = not C_PetJournal.IsFlagFiltered(flag)
            C_PetJournal.SetFlagFilter(flag, value)
        end
        
        local ntypes = C_PetJournal.GetNumPetTypes()
        for i=1,ntypes do
            type_filters[i] = not C_PetJournal.IsPetTypeFiltered(i)
        end
        C_PetJournal.AddAllPetTypesFilter()
        
        local nsources = C_PetJournal.GetNumPetSources()
        for i=1,nsources do
            source_filters[i] = not C_PetJournal.IsPetSourceFiltered(i)
        end
        C_PetJournal.AddAllPetSourcesFilter()
    end

    function filter_load_and_restore()
        if s_search_filter then
            C_PetJournal.SetSearchFilter(s_search_filter)
        end
        
        for flag, value in pairs(flag_filters) do
            C_PetJournal.SetFlagFilter(flag, value)
        end
        
        for flag,value in pairs(type_filters) do
            C_PetJournal.SetPetTypeFilter(flag, value)
        end
        
        for flag,value in pairs(source_filters) do
            C_PetJournal.SetPetSourceFilter(flag, value)
        end
        
        pet_journal_thaw()
    end
end

--
--
--

function lib:LoadPets()
    local petsbyid = self.PetsById

    wipe(self.Pets)
    for pid, p in pairs(petsbyid) do
        petsbyid[pid].state = old
    end
    
    -- save and clear PJ filters
    lib.event_frame:UnregisterEvent("PET_JOURNAL_LIST_UPDATE")
    filter_clear_and_save()
    
    -- scan pets
    local total, owned = C_PetJournal.GetNumPets(false)
    self.NumPets = owned
    self.NumUnknownPets = total - owned
    for i = 1,total do
        local petID, speciesID, isOwned, customName, level,
            favorite, isRevoked, name, icon, petType,
            creatureID, sourceText, description, isWildPet = C_PetJournal.GetPetInfoByIndex(i, false)
        
        if name == nil then
            -- not ready
            self.NumPets = 0
            self.NumUnknownPets = 0
            start_background()
            return
        end
        
        local list, obj
        if isOwned then
            list = self.Pets
            if petsbyid[petid] then
                obj = petsbyid[petid]
                obj.state = "existing"
            else
                obj = { state = "new" }
                petsbyid[petID] = obj
            end
        else
            list = self.UnknownPets
            obj = {}
        end

        obj.petID, obj.speciesID, obj.isOwned, obj.customName, obj.level,
            obj.favorite, obj.isRevoked, obj.name, obj.icon, obj.petType,
            obj.creatureID, obj.sourceText, obj.description, obj.isWildPet 
            = petID, speciesID, isOwned, customName, level,
            favorite, isRevoked, name, icon, petType,
            creatureID, sourceText, description, isWildPet
        obj.isSummonable = C_PetJournal.PetIsSummonable(petID)
        
        tinsert(list, obj)
    end
    
    -- restore PJ filters
    filter_load_and_restore()
       
    -- Signal
    for pid, p in pairs(petsbyid) do
        if petsbyid[pid].state == "old" then
            self.callbacks:Fire("PetRemoved", p)
            petsbyid[pid] = nil
        elseif petsbyid[pid].state == "new" then
            self.callbacks:Fire("PetAdded", p)
        end
    end
    self.callbacks:Fire("PetsUpdated", self)
    
    filter_changed = false
    self.event_frame:Hide()
    lib.event_frame:RegisterEvent("PET_JOURNAL_LIST_UPDATE")
end

lib.event_frame = lib.event_frame or CreateFrame("FRAME")
lib.event_frame:RegisterEvent("COMPANION_UPDATE")
lib.event_frame:RegisterEvent("COMPANION_LEARNED")
lib.event_frame:RegisterEvent("COMPANION_UNLEARNED")
lib.event_frame:RegisterEvent("PET_JOURNAL_LIST_UPDATE")
lib.event_frame:RegisterEvent("ADDON_LOADED")
lib.event_frame:SetScript("OnEvent", function(frame, event, ...)
    if event == "COMPANION_UPDATE" then
        local ctype = ...
        if ctype == nil or ctype == "CRITTER" then
            lib:LoadPets()
        end
    elseif event == "PET_JOURNAL_LIST_UPDATE" then
        start_background()
    elseif event == "ADDON_LOADED" then
        -- we don't care who or what
        lib.event_frame:UnregisterEvent("ADDON_LOADED")
        lib:LoadPets()
    else
        lib:LoadPets()
    end
end)

local timer = 0
function start_background()
    lib.event_frame:Show()
    timer = 10  -- run immediately on next OnUpdate
end

lib.event_frame:SetScript("OnUpdate", function(frame, elapsed)
    timer = timer + elapsed
    if timer > 2 and not InCombatLockdown() then
        if filter_changed then
            filter_changed = false
            if lib.NumUnknownPets > 0 or lib.NumPets > 0 then
                frame:Hide()
            end
            return
        end
        
        lib:LoadPets()
        timer = 0
    end
end)
